unit passhader;
// Lars Middendorf,lmid@gmx.de

interface
uses classes,sysutils;

type


TTokenType=(tkident,tkinteger,tkfloat,tkstring,tkeof,tkerror,
            tkprogram,tkunit,tkinterface,tkimplementation,tkuses,tkwith,tkdo,tkdiscard,tkprocedure,tkfunction,tkforward,tkbegin,tkend,tkasm,tkif,tkthen,tkelse,tkvar,tkvarying,tkparam,tktype,tkrecord,tkidiv,tkmod,tkshr,tkshl,tkand,tkor,tkxor,tknot,
            tkassignment,tkadd,tksub,tkmult,tkdiv,tkequal,tklequal,tkless,tkgequal,tkgreater,tknotequal,tkaddress,tkderef,tksemicolon,tkcomma,tkopenbracket,tkclosebracket,tkpoint,tkdblpoint,tkarrayopen,tkarrayclose,tkrange,
            tktypecast,tklast);
const

tokentext:array[tkident..tklast] of string =('','','','','','',
            'program','unit','interface','implementation','uses','with','do','discard','procedure','function','forward','begin','end','asm','if','then','else','var','varying','param','type','record','div','mod','shr','shl','and','or','xor','not',
            ':=','+','-','*','/','=','<=','<','>=','>','<>','@','^',';',',','(',')','.',':','[',']','..',
            '','');

type
TTokenSet=set of TTokenType;

ESyntaxError=class(Exception);

TScanner=class
private
 current_char,next_char:char;
 stream:TStream;
 function ReadChar:char;
 procedure ReadText;
 procedure FindTokenStart;
 procedure ReadSymbol;
 procedure ReadNum;
 procedure ReadHexNum;
 function ReadString:string;
public
 SourceLine:integer;
 SourcePos:integer;
 TokenString:string;
 TokenType:TTokenType;
 ParsedText:string;
 procedure CheckToken(token:TTokenSet);
 procedure AcceptToken(token:TTokentype);
 function ReadToken:string;
 constructor create(st:TStream);
 destructor destroy;override;
end;

type

TSymbolScope=(scvar,scparam,scattrib,scoutput);

TSymbolInfo=class
 name:string;
 scope:TSymbolScope;
 arraysize:integer;
 offset:integer;
 temp,_default:boolean;
 varying:boolean;
 used:boolean;
end;

TTokenInfo=record
 token:TTokenType;
 SourcePos,SourceLine:integer;
 tokenstring:string;
end;

TExpression=class;
TStatement=class;

TExpression=class
 info:TTokenInfo;
 ident:string;
 subnodes:array[0..1] of TExpression;
 paramlist:array of TExpression;
 symbolinfo:TSymbolInfo;
 sat:boolean;
 unary:boolean;
 lvalue:boolean;
end;

TStatementType=(stblock,stassignment,stwith,stdiscard,stif);

TStatement=class
 info:TTokenInfo;
 _type:TStatementType;
 exp1,exp2:TExpression;
 statements:array of TStatement;
 withdecl:string;
end;


TStack=class
private
 fcount:integer;
 flist:TList;
public
 property Count:Integer read fcount;
 procedure Push(p:pointer);
 function Pop:Pointer;
 function Peek:Pointer;
 constructor create;
 destructor destroy;override;
end;

TARBCommand=record
 cmd:string;
 p1,p2,p3,p4:string;
end;

TAsmOpCode=record
 cmd:string;
 paramcount:integer;
 scalar:boolean;
end; 

TPascalShaderClass=class of TPascalShader;

TPascalShader=class
private
 symbollist,nodelist:TList;
 expstack:TStack;
 exptree:TExpression;
 sc:TScanner;
 codeblock:TStatement;
 tempvarlist:array of string;
 paramsize:integer;
 matrixlist:TStringList;
 withlist:TStringList;
 fsource:string;
 cmdlist:array of TARBCommand;
 reservedwordlist:TStringList;
 SavedDecimalSeparator:char;
 currentcondition,currentconditioninv:string;
 function ReadTokenInfo:TTokenInfo;
 function TokenInfoStr(const ti:TTokenInfo):string;
 function GetOperatorPrecedence(exp:TExpression):integer;
 function CompareOperator(exp1,exp2:TExpression):boolean;
 procedure PushExpression(e:TExpression);
 procedure AddBinaryOperator(e:TExpression);
 procedure AddUnaryOperator(e:TExpression);
 procedure AddOperator(e:TExpression);
 function ParseExpression_r:TExpression;
 function ParseExpression:TExpression;
 function ParseExpressionStatement:TStatement;
 function ParseStatement:TStatement;
 procedure ParseProgram;
 procedure ParseVarBlock;
 procedure ParseVarDecl;
 procedure ParseParamBlock;
 procedure ParseParamDecl;
 procedure ParseVaryingBlock;
 procedure ParseVaryingDecl;
 function ParseWith:TStatement;
 function ParseDiscard:TStatement;
 function Parseif:TStatement; 
 function ParseCodeBlock:TStatement;
 function GetSwizzleMask(const s:string):string;
 function IsValidRSwizzle(const s:string):boolean;
 function IsValidLSwizzle(const s:string):boolean;
 function CorrectSwizzle(const s:string):string;
 function ChangeScalarSwizzle(const s:string):string;
 function CombineSwizzle(const s,sw:string):string;
 function ExtractVarname(const s:string):string;
 function ChangeVarname(const s,n:string):string;
 function FindValidMatrix(const s:string):string;
 function IsValidWithDecl(const s:string):boolean;
 function FindValidWithDecl(const s:string):string;
 procedure CheckIdent(e:TExpression);
 procedure CheckParamCount(e:TExpression;c:integer);
 procedure WriteLine(const s:string);
 procedure WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');overload;
 function NegateStr(const s:string):string;
 function OptimizeExpression(e:TExpression):TExpression;
 function WriteBinaryOperator(exp:TExpression;const target:string=''):string;
 function WriteUnaryOperator(exp:TExpression;const target:string=''):string;
 function WriteOperator(exp:TExpression;const target:string=''):string;
 function WriteExpressionIdent(exp:TExpression;const target:string=''):string;
 function WriteExpressionConst(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_OpCode1(exp:TExpression;const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_OpCode1s(exp:TExpression;const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_OpCode2(exp:TExpression;const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_OpCode2s(exp:TExpression;const opcode:string;const target:string=''):string;
 function WriteExpressionFunc_Normalize(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Length(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_SqrLength(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_InvLength(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Dist(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_InvDist(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_SqrDist(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Sqr(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Sqrt(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Mad(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_MatrixMult3(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_MatrixMult4(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Vector(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_sat(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_clamp(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_reflect(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_lightatten(exp:TExpression;const target:string=''):string;
 function WriteExpression(exp:TExpression;target:string=''):string;
 function Sat(e:TExpression):string;
 function FindSymbol(const sname:string):TSymbolInfo;
 procedure WriteAssignment(st:TStatement);
 procedure WriteStatement(st:TStatement);
 procedure WriteCodeBlock(st:TStatement);
 procedure WriteWithStatement(st:TStatement);
 procedure WriteIfStatement(st:TStatement);
 procedure WriteDiscard(st:TStatement);
 procedure WriteProgram;
 procedure WriteVarBlock;
 procedure WriteVaryingBlock;
 procedure WriteParamBlock;
 procedure WriteCommandList;
 function RemoveDblAssignments:boolean;
 function RemoveAssignments:boolean;
 procedure FindUsedSymbols;
 procedure OptimizeCommandList;
 function CanVarDiscarded(const vn:string;cmdid:integer):boolean;
 function IsVarRead(const cmd:TARBCommand;const v:string):boolean;
 function IsVarWritten(const cmd:TARBCommand;const v:string):boolean;
 procedure AddDefaultSymbol(const s:string;sc:TSymbolScope);
 function AllocateTempVar:string;
 procedure FreeTempVar(const s:string);
 procedure InitMatrixList;
 procedure InitReservedWords;
protected
 AllowDiscard:Boolean;
 SAT_Cmd:boolean;
 VaryingType:TSymbolScope;
 Header:String;
 procedure InitDefaultVars;virtual;
 function WriteExpressionFunction(exp:TExpression;const target:string=''):string;virtual;
public
 property Source:string read fsource;
 function GetSymbolOffset(const v:string):integer;
 constructor create(st:TStream);virtual;
 destructor destroy;override;
end;

TVertexProgram=class(TPascalShader)
protected
 function WriteExpressionFunc_Cos(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_Sin(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_lerp(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunction(exp:TExpression;const target:string=''):string;override;
 procedure InitDefaultVars;override;
public
 constructor create(st:TStream);override;
end;

TFragmentProgram=class(TPascalShader)
protected
 function WriteExpressionFunc_texture(exp:TExpression;const texture:string;const target:string=''):string;
 function WriteExpressionFunc_textureproj(exp:TExpression;const texture:string;const target:string=''):string;
 function WriteExpressionFunc_normalmap(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunc_lerp(exp:TExpression;const target:string=''):string;
 function WriteExpressionFunction(exp:TExpression;const target:string=''):string;override;
 procedure InitDefaultVars;override;
public
 constructor create(st:TStream);override;
end;

implementation

const
binaryop=[tkadd,tksub,tkmult,tkdiv,tkequal,tklequal,tkless,tkgequal,tkgreater,tknotequal,tkpoint,tkand,tkor,tkxor];
unaryop=[tkadd,tksub,tknot];
expressionstart=[tkident,tkinteger,tkfloat]+unaryop;


// TStack

procedure TStack.Push(p:pointer);
begin
inc(fcount);
flist.Add(p);
end;

function TStack.Pop:Pointer;
begin
if count=0 then
 result:=nil
else
begin
 dec(fcount);
 result:=flist.Items[count];
 flist.Delete(count);
end;
end;

function TStack.Peek:Pointer;
begin
if count=0 then
 result:=nil
else
 result:=flist.Items[count-1];
end;

constructor TStack.create;
begin
inherited;
flist:=TList.create;
end;

destructor TStack.destroy;
begin
flist.free;
inherited;
end;


// TScanner

function TScanner.ReadChar:char;
begin
current_char:=next_char;
if stream.Position>=stream.Size then
 next_char:=#0
else
begin
 stream.Read(next_char,1);
 ParsedText:=ParsedText+next_char;
 inc(SourcePos);
end;
if current_char=#10 then inc(sourceline);
result:=current_char;
end;

procedure TScanner.FindTokenStart;
begin
while current_char in [' ',#13,#10] do ReadChar;
if current_char='{' then
begin
 while not (ReadChar in ['}',#0]) do;
end;
end;

procedure TScanner.ReadText;
var
 t:TTokenType;
begin
TokenType:=tkident;
TokenString:=current_char;
while ReadChar in ['A'..'Z','a'..'z','_','0'..'9'] do TokenString:=TokenString+current_char;
for t:=tkident to tklast do
 if sametext(Tokentext[t],TokenString) then
 begin
  TokenType:=t;
  break;
 end;
end;

procedure TScanner.ReadSymbol;
begin
TokenString:=current_char;
TokenType:=tkerror;
case current_char of
 '+':TokenType:=tkadd;
 '-':TokenType:=tksub;
 '*':TokenType:=tkmult;
 '/':
  begin
  if next_char='/' then
  begin
   ReadChar;
   TokenString:='//';
  end;

  TokenType:=tkdiv;
  end;
 ';':TokenType:=tksemicolon;
 ',':TokenType:=tkcomma;
 '(':TokenType:=tkopenbracket;
 ')':TokenType:=tkclosebracket;
 '.':
  begin
   if next_char='.' then
   begin
    TokenType:=tkrange;
    ReadChar;
    TokenString:='..';
   end
   else
   begin
    TokenType:=tkpoint;
   end;
  end;
 '=':TokenType:=tkequal;
 '@':TokenType:=tkaddress;
 '^':TokenType:=tkderef;
 '[':TokenType:=tkarrayopen;
 ']':TokenType:=tkarrayclose;
 '<':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:='<=';
    TokenType:=tklequal;
   end
   else
   if next_char='>' then
   begin
    ReadChar;
    TokenString:='<>';
    TokenType:=tknotequal;
   end
   else
   begin
    TokenType:=tkless;
   end;
  end;
 '>':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:='>=';
    TokenType:=tkgequal;
   end
   else
   begin
    TokenType:=tkgreater;
   end;
  end;
 ':':
  begin
   if next_char='=' then
   begin
    ReadChar;
    TokenString:=':=';
    TokenType:=tkassignment;
   end
   else
   begin
    TokenType:=tkdblpoint;
   end;
  end;
end;
ReadChar;
end;


procedure TScanner.ReadNum;
begin
Tokenstring:=current_char;
TokenType:=tkinteger;
while (ReadChar in ['0'..'9','.']) do
begin
 if (current_char='.') and (next_char='.') then break;
 if (current_char='.') then
 begin
  if TokenType=tkinteger then tokentype:=tkfloat else tokentype:=tkerror;
 end;
 Tokenstring:=Tokenstring+current_char;
end;

end;

procedure TScanner.ReadHexNum;
begin
Tokenstring:=current_char;
TokenType:=tkinteger;
while (ReadChar in ['0'..'9','A'..'F','a'..'f']) do
begin
 Tokenstring:=Tokenstring+current_char;
end;

end;


function TScanner.ReadString:string;
begin
while not (ReadChar in ['''',#13,#10,#0]) do TokenString:=TokenString+current_char;
if current_char='''' then TokenType:=tkstring else TokenType:=tkerror;
ReadChar;
end;

function TScanner.ReadToken:string;
begin
FindTokenStart;
TokenString:='';
TokenType:=tkerror;
case current_char of
 'A'..'Z','a'..'z','_':ReadText;
 '0'..'9':ReadNum;
 '$':ReadHexNum;
 '''':ReadString;
 #0:
 begin
  result:=#0;
  TokenType:=tkeof;
 end
else
 ReadSymbol;
if TokenString='//' then
begin
while not (ReadChar in [#13,#10,#0]) do;
TokenString:=ReadToken;
end;
end;
result:=TokenString;
end;

procedure TScanner.CheckToken(token:TTokenSet);
begin
if not (TokenType in token) then raise ESyntaxError.Create('Invalid Token');
end;

procedure TScanner.AcceptToken(token:TTokentype);
begin
CheckToken([token]);
ReadToken;
end;


constructor TScanner.create(st:TStream);
begin
inherited create;
SourceLine:=1;
SourcePos:=2;
stream:=st;
stream.ReadBuffer(current_char,1);
stream.ReadBuffer(next_char,1);
ParsedText:=current_char+next_char;
ReadToken;
end;

destructor TScanner.destroy;
begin
inherited;
end;

// TPascalShader

function TPascalShader.ReadTokenInfo:TTokenInfo;
begin
result.token:=sc.TokenType;
result.SourcePos:=sc.SourcePos;
result.SourceLine:=sc.SourceLine;
result.tokenstring:=sc.TokenString;
end;

function TPascalShader.TokenInfoStr(const ti:TTokenInfo):string;
begin
 result:='Line '+inttostr(ti.SourceLine)+':';
end;

function TPascalShader.GetOperatorPrecedence(exp:TExpression):integer;
begin
result:=-1;
case exp.info.token of
 tkpoint:result:=5;
 tknot,tkaddress:result:=4;
 tkmult,tkdiv,tkidiv,tkmod,tkand,tkshl,tkshr:result:=3;
 tkadd,tksub,tkor,tkxor:if exp.unary then result:=4 else result:=2;
 tkequal,tknotequal,tklequal,tkless,tkgreater,tkgequal:result:=1;
end;
end;

function TPascalShader.CompareOperator(exp1,exp2:TExpression):boolean;
begin
result:=GetOperatorPrecedence(exp1)>=GetOperatorPrecedence(exp2);
end;

procedure TPascalShader.PushExpression(e:TExpression);
begin
expstack.Push(e);
exptree:=e;
end;



procedure TPascalShader.ParseVarDecl;
var
 si:TSymbolInfo;
begin
repeat
 if (FindSymbol(sc.TokenString)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1) then
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Identifier redeclared: "'+sc.TokenString+'"');
 si:=TSymbolInfo.Create;
 si.name:=sc.TokenString;
 si.scope:=scvar;
 symbollist.Add(si);
 sc.AcceptToken(tkident);
sc.CheckToken([tkcomma,tksemicolon]);
if sc.tokentype=tkcomma then sc.AcceptToken(tkcomma);
until sc.TokenType=tksemicolon;
sc.AcceptToken(tksemicolon);
end;

procedure TPascalShader.ParseVarBlock;
begin
sc.AcceptToken(tkvar);
sc.CheckToken([tkident,tkbegin,tkparam,tkvar,tkvarying]);
while sc.TokenType=tkident do
begin
 case sc.TokenType of
  tkIdent:
   begin
    ParseVarDecl;
   end;
 end;
end;

end;

procedure TPascalShader.ParseParamDecl;
var
 si,si2:TSymbolInfo;
 i,c:integer;
begin
repeat
 if (FindSymbol(sc.TokenString)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1) then
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Identifier redeclared: "'+sc.TokenString+'"');;
 si:=TSymbolInfo.Create;
 si.name:=sc.TokenString;
 si.scope:=scparam;
 sc.AcceptToken(tkident);
 sc.CheckToken([tkcomma,tksemicolon,tkarrayopen]);
 if sc.tokentype=tkarrayopen then
 begin
  sc.AcceptToken(tkarrayopen);
  sc.CheckToken([tkinteger]);
  c:=strtoint(sc.tokenstring);
  sc.AcceptToken(tkinteger);
  sc.AcceptToken(tkarrayclose);
  for i:=0 to c-1 do
  begin
   si2:=TSymbolInfo.Create;
   if i=0 then
    si2.arraysize:=c
   else
    si2.arraysize:=-1;   
   si2.offset:=paramsize+i;
   si2.name:=si.name+'['+inttostr(i)+']';
   si2.scope:=si.scope;
   symbollist.add(si2);
  end;
  inc(paramsize,c);
  si.free;
 end
 else
  begin
  symbollist.add(si);
  si.offset:=paramsize;
  inc(paramsize);
  end;
sc.CheckToken([tkcomma,tksemicolon]);
if sc.tokentype=tkcomma then sc.AcceptToken(tkcomma);
until sc.TokenType=tksemicolon;
sc.AcceptToken(tksemicolon);
end;

procedure TPascalShader.ParseParamBlock;
begin
sc.AcceptToken(tkparam);
sc.CheckToken([tkident,tkbegin,tkvar,tkparam,tkvarying]);
while sc.TokenType=tkident do
begin
 case sc.TokenType of
  tkIdent:
   begin
    ParseParamDecl;
   end;
 end;
end;

end;


procedure TPascalShader.ParseVaryingDecl;
var
 si:TSymbolInfo;
begin
repeat
 if (FindSymbol(sc.TokenString)<>nil) or (reservedwordlist.IndexOf(sc.TokenString)<>-1) then
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Identifier redeclared: "'+sc.TokenString+'"');
 si:=TSymbolInfo.Create;
 si.name:=sc.TokenString;
 si.scope:=VaryingType;
 si.varying:=true;
 symbollist.Add(si);
 sc.AcceptToken(tkident);
sc.CheckToken([tkcomma,tksemicolon]);
if sc.tokentype=tkcomma then sc.AcceptToken(tkcomma);
until sc.TokenType=tksemicolon;
sc.AcceptToken(tksemicolon);
end;

procedure TPascalShader.ParseVaryingBlock;
begin
sc.AcceptToken(tkvarying);
sc.CheckToken([tkident,tkbegin,tkparam,tkvar,tkvarying]);
while sc.TokenType=tkident do
begin
 case sc.TokenType of
  tkIdent:
   begin
    ParseVaryingDecl;
   end;
 end;
end;

end;


procedure TPascalShader.AddBinaryOperator(e:TExpression);
begin
if expstack.count<2 then
begin
 raise ESyntaxError.Create(TokenInfoStr(e.info)+'Expression expected but "'+sc.TokenString+'" found');
end;
e.subnodes[1]:=TExpression(expstack.pop);
e.subnodes[0]:=TExpression(expstack.pop);
PushExpression(e);
exptree:=e;
end;

procedure TPascalShader.AddUnaryOperator(e:TExpression);
begin
if expstack.count<1 then
begin
 raise ESyntaxError.Create(TokenInfoStr(e.info)+'Expression expected but "'+sc.TokenString+'" found');
end;
e.subnodes[0]:=TExpression(expstack.pop);
PushExpression(e);
exptree:=e;
end;

procedure TPascalShader.AddOperator(e:TExpression);
begin
if e.unary then AddUnaryOperator(e) else AddBinaryOperator(e);
end;


function TPascalShader.ParseExpression_r:TExpression;
const
 endtoken=[tkcomma,tksemicolon,tkassignment,tkend,tkelse,tkthen];
var
 e:TExpression;
 preoperator:boolean;
 opstack:TStack;
 l:integer;
 b:boolean;
begin
sc.CheckToken([tkident,tkinteger,tkfloat,tkopenbracket]+unaryop);
preoperator:=true;
opstack:=TStack.Create;
repeat
case sc.TokenType of
 tkident:
  begin
   e:=TExpression.create;
   e.info:=ReadTokenInfo;
   nodelist.Add(e);
   repeat
   b:=true;
   e.ident:=e.ident+sc.TokenString;
   sc.AcceptToken(tkident);
   if sc.TokenType=tkarrayopen then
   begin
    e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkarrayopen);
    e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkinteger);
    e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkarrayclose);
   end;
   if sc.TokenType=tkpoint then
   begin
   e.ident:=e.ident+sc.TokenString;
    sc.AcceptToken(tkpoint);
    sc.CheckToken([tkident]);
    b:=false;
   end;
   until b;
   if sc.TokenType=tkopenbracket then
   begin
    sc.AcceptToken(tkopenbracket);
    while sc.TokenType<>tkclosebracket do
    begin
     l:=length(e.paramlist);
     setlength(e.paramlist,l+1);
     e.paramlist[l]:=ParseExpression;
     sc.CheckToken([tkcomma,tkclosebracket]);
     if sc.TokenType=tkcomma then sc.AcceptToken(tkcomma);
    end;
    sc.AcceptToken(tkclosebracket);
   end;
   PushExpression(e);
   sc.CheckToken(binaryop+endtoken+[tkarrayclose,tkclosebracket]);
   preoperator:=false;
  end;
 tkinteger,tkfloat:
  begin
   e:=TExpression.create;
   e.info:=ReadTokenInfo;
   nodelist.Add(e);
   PushExpression(e);
   sc.AcceptToken(sc.TokenType);
   sc.CheckToken(binaryop+endtoken+[tkclosebracket]);
   preoperator:=false;
  end;
  tkadd,tksub,tkmult,tkdiv,tkequal,
  tklequal,tkless,tkgequal,tkgreater,tknotequal,
  tkand,tkor,tkxor:
  begin
   e:=TExpression.Create;
   e.info:=ReadTokenInfo;
   e.unary:=preoperator or (sc.TokenType in [tknot]);
   nodelist.Add(e);
   sc.AcceptToken(e.info.token);
   while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
   begin
    AddOperator(TExpression(opstack.Pop));
   end;
   opstack.push(e);
   if preoperator then
   begin
    sc.CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkopenbracket]-[tkadd,tksub]);
    preoperator:=false;
   end
   else
   begin
    sc.CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkopenbracket]);
    preoperator:=true;
   end;
  end;
 tkpoint:
 begin
  e:=TExpression.Create;
  e.info:=ReadTokenInfo;
  e.unary:=false;
  nodelist.Add(e);
  sc.AcceptToken(e.info.token);
  while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
  begin
   AddOperator(TExpression(opstack.Pop));
  end;
  opstack.push(e);
  sc.CheckToken([tkident]);
  preoperator:=false;
 end;
 tkopenbracket:
 begin
  sc.AcceptToken(tkopenbracket);
  ParseExpression_r;
  sc.AcceptToken(tkclosebracket);
  preoperator:=false;
  sc.CheckToken(binaryop+endtoken+[tkclosebracket,tkderef]);
 end;
 tknot:
 begin
   e:=TExpression.Create;
   e.info:=ReadTokenInfo;
   e.unary:=true;
   nodelist.Add(e);
   sc.AcceptToken(e.info.token);
   while (opstack.count>0) and CompareOperator(TExpression(opstack.Peek),e) do
   begin
    AddOperator(TExpression(opstack.Pop));
   end;
   opstack.push(e);
   sc.CheckToken(unaryop+[tkident,tkinteger,tkfloat,tkopenbracket]);
   preoperator:=true;
 end;
end;
until sc.TokenType in [tksemicolon,tkcomma,tkassignment,tkclosebracket,tkend,tkelse,tkthen];

while (opstack.count>0) do
begin
 AddOperator(TExpression(opstack.Pop));
end;
result:=exptree;
opstack.free;
end;

function TPascalShader.ParseExpression:TExpression;
var
 start:integer;
begin
start:=expstack.count;
result:=ParseExpression_r;
if expstack.count-start<>1 then
 raise ESyntaxError.Create(TokenInfoStr(result.info)+'Expression expected but "'+sc.TokenString+'" found');
expstack.pop;
end;

function TPascalShader.FindSymbol(const sname:string):TSymbolInfo;
var
 i,j:integer;
begin
result:=nil;
for j:=withlist.count-1 downto 0 do
begin
 for i:=0 to symbollist.count-1 do
  if sametext(TSymbolInfo(symbollist[i]).name,withlist[j]+'.'+sname) then
  begin
   result:=TSymbolInfo(symbollist[i]);
   break;
  end;
 if result<>nil then break;
end;
if result=nil then
for i:=0 to symbollist.count-1 do
 if sametext(TSymbolInfo(symbollist[i]).name,sname) then
 begin
  result:=TSymbolInfo(symbollist[i]);
  break;
 end;
end;

function TPascalShader.GetSwizzleMask(const s:string):string;
var
 p:integer;
begin
p:=pos('.',s);
if (p=0) or assigned(findsymbol(s)) or (pos('}',ExtractFileExt(s))<>0) then
begin
result:='';
exit;
end;
result:=ExtractFileExt(s);
result:=copy(result,2,length(result)-1);
end;

function TPascalShader.IsValidRSwizzle(const s:string):boolean;
var
 a,b,i:integer;
begin
a:=0;
b:=0;
for i:=1 to length(s) do
case Upcase(s[i]) of
 'X','Y','Z','W':inc(a);
 'R','G','B','A':inc(b);
end;
result:=(a+b) in [1,4];
end;

function TPascalShader.IsValidLSwizzle(const s:string):boolean;
var
 a,b,c,d,i:integer;
begin
a:=0;
b:=0;
c:=0;
d:=0;
for i:=1 to length(s) do
case UpCase(s[i]) of
 'X','R':inc(a);
 'Y','G':inc(b);
 'Z','B':inc(c);
 'W','A':inc(d);
end;
result:=((a<=1) and (b<=1) and (c<=1) and (d<=1)) and (a+b+c+d>0);
end;

function TPascalShader.CorrectSwizzle(const s:string):string;
var
 sw,v:string;
begin
sw:=GetSwizzleMask(s);
if sw='' then
begin
 result:=s;
end
else
begin
sw:=StringReplace(sw,'r','x',[rfreplaceall]);
sw:=StringReplace(sw,'g','y',[rfreplaceall]);
sw:=StringReplace(sw,'b','z',[rfreplaceall]);
sw:=StringReplace(sw,'a','w',[rfreplaceall]);
v:=copy(s,1,length(s)-length(sw)-1);
result:=v+'.'+sw;
end;
end;

function TPascalShader.ChangeScalarSwizzle(const s:string):string;
var
 v,sw:string;
begin
sw:=GetSwizzleMask(s);
if sw='' then
 v:=s
else
 v:=copy(s,1,length(s)-length(sw)-1);
if sw='' then
 result:=v+'.x'
else
 result:=v+'.'+sw[1];
end;

function TPascalShader.CombineSwizzle(const s,sw:string):string;
var
 v,sw2,sw3:string;
 i:integer;
begin
sw2:=lowercase(GetSwizzleMask(s));
if sw2='' then
 v:=s
else
v:=copy(s,1,length(s)-length(sw2)-1);
if sw2='' then sw2:='xyzw';
 sw3:='';
for i:=1 to length(sw) do
 if pos(lowercase(sw[i]),sw2)<>0 then
  sw3:=sw3+lowercase(sw[i]);
if sw3='' then
 result:=''
else
 result:=v+'.'+sw3;
end;

function TPascalShader.ExtractVarname(const s:string):string;
var
 sw:string;
begin
sw:=lowercase(GetSwizzleMask(s));
if sw='' then
 result:=s
else
 result:=copy(s,1,length(s)-length(sw)-1);
if copy(result,1,1)='-' then
 result:=copy(result,2,length(result)-1);
end;

function TPascalShader.ChangeVarname(const s,n:string):string;
var
 sw,p:string;
begin
sw:=lowercase(GetSwizzleMask(s));
if sw<>'' then sw:='.'+sw;
if copy(s,1,1)='-' then p:='-' else p:='';
result:=p+n+sw;
end;

procedure TPascalShader.CheckIdent(e:TExpression);
var
 v,s:string;
 b:boolean;
begin
if e.info.token=tkident then
begin
 e.ident:=CorrectSwizzle(e.ident);
 s:=GetSwizzleMask(e.ident);
 if s='' then
  v:=e.ident
 else
  v:=copy(e.ident,1,length(e.ident)-length(s)-1);
 e.symbolinfo:=FindSymbol(v);
 if e.symbolinfo=nil then
  raise ESyntaxError.Create(TokenInfoStr(e.info)+'Undeclared Identifier: "'+e.ident+'"');
 if s='' then
  e.ident:=e.symbolinfo.name
 else
 e.ident:=e.symbolinfo.name+'.'+s;
 if not e.lvalue and (e.symbolinfo.scope=scoutput) then
  raise ESyntaxError.Create(TokenInfoStr(e.info)+'Variable ist writeonly: "'+e.ident+'"');
 if s<>'' then
 begin
  if e.lvalue then b:=IsValidLSwizzle(s) else b:=IsValidRSwizzle(s);
  if not b then
   raise ESyntaxError.Create(TokenInfoStr(e.info)+'Invalid swizzle: "'+s+'"');
 end;
end
else
 raise ESyntaxError.Create(TokenInfoStr(e.info)+'Identifier expected ');
end;

function TPascalShader.FindValidMatrix(const s:string):string;
var
 i:integer;
begin
result:='';
 for i:=withlist.count-1 downto 0 do
  if matrixlist.IndexOf(withlist[i]+'.'+s)<>-1 then
  begin
   result:=withlist[i]+'.'+s;
   break;
  end;
if (result='') and (matrixlist.IndexOf(s)<>-1) then
 result:=s;
end;

function TPascalShader.IsValidWithDecl(const s:string):boolean;
var
 i:integer;
begin
result:=false;
for i:=0 to matrixlist.count-1 do
 if sametext(s+'.',copy(matrixlist[i],1,length(s)+1)) then
 begin
  result:=true;
  break;
 end;
if not result then
for i:=0 to symbollist.count-1 do
 if sametext(s+'.',copy(TSymbolInfo(symbollist[i]).name,1,length(s)+1)) then
 begin
  result:=true;
  break;
 end;

end;

function TPascalShader.FindValidWithDecl(const s:string):string;
var
 i:integer;
begin
result:='';
 for i:=withlist.count-1 downto 0 do
  if IsValidWithDecl(withlist[i]+'.'+s) then
  begin
   result:=withlist[i]+'.'+s;
   break;
  end;
if (result='') and IsValidWithDecl(s) then
 result:=s;

end;


procedure TPascalShader.CheckParamCount(e:TExpression;c:integer);
begin
if length(e.paramlist)<>c then
 raise ESyntaxError.Create(tokeninfostr(e.info)+'Invalid parameter count');
end;

function TPascalShader.ParseExpressionStatement:TStatement;
begin
result:=TStatement.Create;
nodelist.Add(result);
result.exp1:=ParseExpression;
result.exp1.lvalue:=true;
sc.CheckToken([tkassignment]);
sc.AcceptToken(tkassignment);
result._type:=stassignment;
result.exp2:=ParseExpression;
//sc.AcceptToken(tksemicolon);
end;

function TPascalShader.ParseCodeBlock:TStatement;
var
 l:integer;
begin
result:=TStatement.Create;
result._type:=stblock;
nodelist.Add(result);
sc.AcceptToken(tkbegin);
sc.CheckToken([tkend,tkident,tkfloat,tkinteger,tkopenbracket,tkwith,tkdiscard,tkif]+expressionstart);
while sc.TokenType<>tkend do
begin
while sc.TokenType=tksemicolon do sc.AcceptToken(tksemicolon);
l:=length(result.statements);
setlength(result.statements,l+1);
result.statements[l]:=ParseStatement;
while sc.TokenType=tksemicolon do sc.AcceptToken(tksemicolon);
sc.CheckToken([tkend,tkident,tkfloat,tkinteger,tkopenbracket,tkwith,tkdiscard,tkif]+expressionstart);
end;
sc.AcceptToken(tkend);
end;

function TPascalShader.ParseWith:TStatement;
var
 b:boolean;
begin
result:=TStatement.Create;
result._type:=stwith;
nodelist.add(result);
result.info:=ReadTokenInfo;
sc.AcceptToken(tkwith);
sc.CheckToken([tkident]);
repeat
b:=true;
result.withdecl:=result.withdecl+sc.TokenString;
sc.AcceptToken(tkident);
if sc.TokenType=tkarrayopen then
begin
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkarrayopen);
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkinteger);
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkarrayclose);
end;
if sc.TokenType=tkpoint then
begin
 result.withdecl:=result.withdecl+sc.TokenString;
 sc.AcceptToken(tkpoint);
 sc.CheckToken([tkident]);
 b:=false;
end;
until b;
sc.AcceptToken(tkdo);
setlength(result.statements,1);
result.statements[0]:=ParseStatement;
end;

function TPascalShader.ParseDiscard:TStatement;
begin
result:=TStatement.Create;
result._type:=stdiscard;
nodelist.add(result);
result.info:=ReadTokenInfo;
if not AllowDiscard then
 raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+' Discard not allowed');
sc.AcceptToken(tkdiscard);
end;

function TPascalShader.Parseif:TStatement;
begin
result:=TStatement.Create;
result._type:=stif;
nodelist.add(result);
result.info:=ReadTokenInfo;
sc.AcceptToken(tkif);
Result.exp1:=ParseExpression;
sc.AcceptToken(tkthen);
setlength(result.statements,1);
result.statements[0]:=ParseStatement;
if sc.TokenType=tkelse then
begin
 sc.AcceptToken(tkelse);
 setlength(result.statements,2);
 result.statements[1]:=ParseStatement;
end;
end;


function TPascalShader.ParseStatement:TStatement;
var
 inf:TTokenInfo;
begin
sc.CheckToken([tkbegin,tkopenbracket,tkwith,tkdiscard,tkif,tkinteger,tkfloat,tkadd,tksub,tknot]+expressionstart);
inf:=ReadTokenInfo;
case sc.TokenType of
 tkbegin:result:=ParseCodeBlock;
 tkident,tkopenbracket,tkinteger,tkfloat,tkadd,tksub,tknot:
  begin
   result:=ParseExpressionStatement;
  end;
 tkwith:
  begin
  result:=ParseWith;
  end;
  tkdiscard:
   begin
   result:=ParseDiscard;
   end;
  tkif:
   begin
    result:=Parseif;
   end;
 else
  raise ESyntaxError.Create(TokenInfoStr(ReadTokenInfo)+'Statement expected');
end;
result.info:=inf;
end;


procedure TPascalShader.ParseProgram;
begin
sc.CheckToken([tkvar,tkparam,tkvarying,tkbegin]);
while sc.TokenType<>tkbegin do
begin
 case sc.TokenType of
  tkvar:ParseVarBlock;
  tkparam:ParseParamBlock;
  tkvarying:ParseVaryingBlock;
 end;
end;
codeblock:=ParseCodeBlock;
end;

procedure TPascalShader.AddDefaultSymbol(const s:string;sc:TSymbolScope);
var
 si:TSymbolInfo;
begin
si:=TSymbolInfo.Create;
si.name:=s;
si.scope:=sc;
si._default:=true;
symbollist.Add(si)
end;

procedure TPascalShader.WriteLine(const s:string);
begin
fSource:=fSource+s+#13#10;
end;


procedure TPascalShader.WriteCmd(const cmd,p1,p2,p3:string;const p4:string='');
var
 l:integer;
begin
l:=length(cmdlist);
setlength(cmdlist,l+1);
cmdlist[l].cmd:=cmd;
cmdlist[l].p1:=p1;
cmdlist[l].p2:=p2;
cmdlist[l].p3:=p3;
cmdlist[l].p4:=p4;
end;

function TPascalShader.NegateStr(const s:string):string;
begin
result:=trim(s);
if copy(s,1,1)='-' then
 result:=copy(result,2,length(result)-1)
else
 result:='-'+result;
end;

function TPascalShader.Sat(e:TExpression):string;
begin
if e.sat then result:='_SAT' else result:='';
end;

function TPascalShader.OptimizeExpression(e:TExpression):TExpression;
var
 i:integer;
begin
result:=e;
if assigned(e.subnodes[0]) then e.subnodes[0]:=OptimizeExpression(e.subnodes[0]);
if assigned(e.subnodes[1]) then e.subnodes[1]:=OptimizeExpression(e.subnodes[1]);
for i:=0 to length(e.paramlist)-1 do
 e.paramlist[i]:=OptimizeExpression(e.paramlist[i]);
case e.info.token of
tkadd:
 begin
  if e.subnodes[0].info.token=tkmult then
  begin
   e.info.token:=tkident;
   e.ident:='mad';
   setlength(e.paramlist,3);
   e.paramlist[0]:=e.subnodes[0].subnodes[0];
   e.paramlist[1]:=e.subnodes[0].subnodes[1];
   e.paramlist[2]:=e.subnodes[1];
  end
  else
  if e.subnodes[1].info.token=tkmult then
  begin
   e.info.token:=tkident;
   e.ident:='mad';
   setlength(e.paramlist,3);
   e.paramlist[0]:=e.subnodes[1].subnodes[0];
   e.paramlist[1]:=e.subnodes[1].subnodes[1];
   e.paramlist[2]:=e.subnodes[0];
  end
  else
  if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (e.subnodes[1].info.token in [tkinteger,tkfloat]) then
  begin
   e.info.token:=tkfloat;
   e.info.tokenstring:=floattostr(strtofloat(e.subnodes[0].info.tokenstring)+strtofloat(e.subnodes[1].info.tokenstring));
  end;
 end;
tkmult:
 begin
  if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (e.subnodes[1].info.token in [tkinteger,tkfloat]) then
  begin
   e.info.token:=tkfloat;
   e.info.tokenstring:=floattostr(strtofloat(e.subnodes[0].info.tokenstring)*strtofloat(e.subnodes[1].info.tokenstring));
  end
  else
  if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(e.subnodes[0].info.tokenstring)-1)<0.00000001) then
  begin
   result:=e.subnodes[1];
  end
  else
  if (e.subnodes[1].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(e.subnodes[1].info.tokenstring)-1)<0.00000001) then
  begin
   result:=e.subnodes[0];
  end
  else
  if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(e.subnodes[0].info.tokenstring))<0.00000001) then
  begin
   e.info.token:=tkinteger;
   e.info.tokenstring:='0';
  end
  else
  if (e.subnodes[1].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(e.subnodes[1].info.tokenstring))<0.00000001) then
  begin
   e.info.token:=tkinteger;
   e.info.tokenstring:='0';
  end;
 end;
tksub:
 begin
  if e.unary then
  begin
   if (e.subnodes[0].info.token in [tkinteger,tkfloat])then
   begin
  //  e.info.token:=tkfloat;
  //  e.info.tokenstring:=floattostr(-strtofloat(e.subnodes[0].info.tokenstring));
   end;
  end
  else
  begin
   if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (e.subnodes[1].info.token in [tkinteger,tkfloat]) then
   begin
    e.info.token:=tkfloat;
    e.info.tokenstring:=floattostr(strtofloat(e.subnodes[0].info.tokenstring)-strtofloat(e.subnodes[1].info.tokenstring));
   end;
  end;
 end;
tkdiv:
 begin
  if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (e.subnodes[1].info.token in [tkinteger,tkfloat]) then
  begin
   e.info.token:=tkfloat;
   e.info.tokenstring:=floattostr(strtofloat(e.subnodes[0].info.tokenstring)/strtofloat(e.subnodes[1].info.tokenstring));
  end
  else
  if (e.subnodes[1].info.token in [tkinteger,tkfloat]) then
  begin
   e.info.token:=tkmult;
   e.subnodes[1].info.tokenstring:=floattostr(1/strtofloat(e.subnodes[1].info.tokenstring));
   result:=OptimizeExpression(e);
  end
  else
  if (e.subnodes[0].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(e.subnodes[0].info.tokenstring))<0.00000001) then
  begin
   e.info.token:=tkinteger;
   e.info.tokenstring:='0';
  end
  else
  if (e.subnodes[1].info.token in [tkinteger,tkfloat]) and (abs(strtofloat(e.subnodes[1].info.tokenstring))<0.00000001) then
  begin
   e.info.token:=tkinteger;
   e.info.tokenstring:='0';
  end
  else
  if sametext(e.subnodes[1].info.tokenstring,'sqrt') then
  begin
   e.info.token:=tkmult;
   e.subnodes[1].info.tokenstring:='rcp';
   e.subnodes[1].ident:='rcp';
   result:=OptimizeExpression(e);
  end
  else
  if sametext(e.subnodes[1].info.tokenstring,'length') then
  begin
   e.info.token:=tkmult;
   e.subnodes[1].info.tokenstring:='invlength';
   e.subnodes[1].ident:='invlength';
   result:=OptimizeExpression(e);
  end
  else
  if sametext(e.subnodes[1].info.tokenstring,'dist') then
  begin
   e.info.token:=tkmult;
   e.subnodes[1].info.tokenstring:='invdist';
   e.subnodes[1].ident:='invdist';
   result:=OptimizeExpression(e);
  end;

 end;
tkident:
 begin
 if SameText(e.ident,'pi') then
 begin
  e.info.token:=tkfloat;
  e.info.tokenstring:='3.1415926535';
 end;
 if SameText(e.ident,'sqr') and (length(e.paramlist)=1) then
  if (e.paramlist[0].info.token=tkident) and SameText(e.paramlist[0].ident,'length') then
  begin
  e.paramlist[0].ident:='sqrlength';
  result:=OptimizeExpression(e.paramlist[0]);
  end;
 if SameText(e.ident,'sqr') and (length(e.paramlist)=1) then
  if (e.paramlist[0].info.token=tkident) and SameText(e.paramlist[0].ident,'dist') then
  begin
  e.paramlist[0].ident:='sqrdist';
  result:=OptimizeExpression(e.paramlist[0]);
  end;
 if SameText(e.ident,'sqrt') and (length(e.paramlist)=1) then
  if (e.paramlist[0].info.token in [tkinteger,tkfloat]) then
  begin
  e.info.token:=tkfloat;
  e.info.tokenstring:=floattostr(sqrt(strtofloat(e.paramlist[0].info.tokenstring)));
  end;
 if SameText(e.ident,'sqr') and (length(e.paramlist)=1) then
  if (e.paramlist[0].info.token in [tkinteger,tkfloat]) then
  begin
  e.info.token:=tkfloat;
  e.info.tokenstring:=floattostr(sqr(strtofloat(e.paramlist[0].info.tokenstring)));
  end;
 if SameText(e.ident,'rsqrt') and (length(e.paramlist)=1) then
  if (e.paramlist[0].info.token in [tkinteger,tkfloat]) then
  begin
  e.info.token:=tkfloat;
  e.info.tokenstring:=floattostr(1/sqrt(strtofloat(e.paramlist[0].info.tokenstring)));
  end;

 if SAT_Cmd and (SameText(e.ident,'sat') and (length(e.paramlist)=1)) then
  begin
  e.paramlist[0].sat:=true;
  result:=OptimizeExpression(e.paramlist[0]);
  end;

 end;
end;
end;

function TPascalShader.AllocateTempVar:string;
var
 si:TSymbolInfo;
 i:integer;
begin
if tempvarlist=nil then
begin
 i:=-1;
 repeat
inc(i);
  si:=FindSymbol('temp'+inttostr(i));
 until si=nil;
 si:=TSymbolInfo.Create;
 si.scope:=scvar;
 si.name:='temp'+inttostr(i);
 si.temp:=true;
 symbollist.add(si);
 result:=si.name;
end
else
begin
 result:=tempvarlist[length(tempvarlist)-1];
 setlength(tempvarlist,length(tempvarlist)-1);
end;
end;

procedure TPascalShader.FreeTempVar(const s:string);
var
 l:integer;
 si:TSymbolInfo;
begin
 si:=FindSymbol(s);
 if (si<>nil) and (si.temp) then
 begin
 l:=length(tempvarlist);
 setlength(tempvarlist,l+1);
 tempvarlist[l]:=si.name;
 end;
end;


function TPascalShader.WriteBinaryOperator(exp:TExpression;const target:string=''):string;
var
 p1,p2,op,temp:string;
begin
p1:=WriteExpression(exp.subnodes[0]);
p2:=WriteExpression(exp.subnodes[1]);
FreeTempVar(p1);
FreeTempVar(p2);
result:='';
case exp.info.token of
 tkadd,tksub,tkmult,tkor,tkand,tkgequal,tkless:
 begin
  case exp.info.token of
   tkadd:op:='ADD';
   tksub:op:='SUB';
   tkmult:op:='MUL';
   tkor:op:='MAX';
   tkand:op:='MIN';
   tkgequal:op:='SGE';
   tkless:op:='SLT';
  end;
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd(op+sat(exp),result,p1,p2);
  end
  else
  begin
   WriteCmd(op+sat(exp),target,p1,p2);
   result:=target;
  end;
 end;
 tkgreater:
 begin
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd('SLT'+sat(exp),result,NegateStr(p1),NegateStr(p2));
  end
  else
  begin
   WriteCmd('SLT'+sat(exp),target,NegateStr(p1),NegateStr(p2));
   result:=target;
  end;
 end;
 tklequal:
 begin
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd('SGE'+sat(exp),result,NegateStr(p1),NegateStr(p2));
  end
  else
  begin
   WriteCmd('SGE'+sat(exp),target,NegateStr(p1),NegateStr(p2));
   result:=target;
  end;
 end;
 tkequal:
 begin
   result:=AllocateTempVar;
   WriteCmd('SUB',result,p1,p2);
   WriteCmd('ABS',result,result,'');
   if target<>'' then
   begin
    freetempvar(result);
    result:=target;
   end;
   WriteCmd('SLT'+sat(exp),result,p1,'0.001');
   end;
 tknotequal,tkxor:
 begin
   result:=AllocateTempVar;
   WriteCmd('SUB',result,p1,p2);
   WriteCmd('ABS',result,result,'');
     if target<>'' then
   begin
    freetempvar(result);
    result:=target;
   end;
   WriteCmd('SGE'+sat(exp),result,p1,'0.001');
   end;
 tkdiv:
 begin
   p2:=ChangeScalarSwizzle(p2);
   result:=AllocateTempVar;
   WriteCmd('RCP',result,p2,'');
   if target<>'' then
   begin
    temp:=result;
    result:=target;
    WriteCmd('MUL'+sat(exp),result,temp,p1);
    FreeTempVar(temp);
   end
   else
    WriteCmd('MUL'+sat(exp),result,result,p1);
  end;
end;
end;

function TPascalShader.WriteUnaryOperator(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
case exp.info.token of
 tksub:
 begin
 if target='' then
 begin
  if exp.sat then
  begin
   p1:=WriteExpression(exp.subnodes[0]);
   FreeTempVar(p1);
   result:=AllocateTempVar;
   WriteCmd('MOV'+sat(exp),result,negatestr(p1),'');
  end
  else
   result:=negatestr(WriteExpression(exp.subnodes[0],target));
 end
 else
 begin
  p1:=WriteExpression(exp.subnodes[0]);
  FreeTempVar(p1);
  WriteCmd('MOV'+sat(exp),target,negatestr(p1),'');
  result:=target;
 end;
 end;
 tkadd:
 begin
 if target='' then
 begin
  if exp.sat then
  begin
   p1:=WriteExpression(exp.subnodes[0]);
   FreeTempVar(p1);
   result:=AllocateTempVar;
   WriteCmd('MOV'+sat(exp),result,p1,'');
  end
  else
   result:=WriteExpression(exp.subnodes[0],target);
 end
 else
 begin
  p1:=WriteExpression(exp.subnodes[0]);
  FreeTempVar(p1);
  WriteCmd('MOV'+sat(exp),target,p1,'');
  result:=target;
 end;
 end;
 tknot:
 begin
  p1:=WriteExpressionIdent(exp.subnodes[0]);
  FreeTempVar(p1);
  if target='' then
  begin
   result:=AllocateTempVar;
   WriteCmd('SUB'+sat(exp),result,'1',p1);
  end
  else
  begin
   WriteCmd('SUB'+sat(exp),target,'1',p1);
   result:=target;
  end;
 end;
end;
end;

function TPascalShader.WriteOperator(exp:TExpression;const target:string=''):string;
begin
if exp.unary then
 result:=WriteUnaryOperator(exp,target)
else
 result:=WriteBinaryOperator(exp,target);
end;

function TPascalShader.WriteExpressionIdent(exp:TExpression;const target:string=''):string;
begin
 CheckIdent(exp);
 if target='' then
 begin
 if exp.sat then
 begin
  result:=AllocateTempVar;
  WriteCmd('MOV'+sat(exp),result,exp.ident,'');
 end
 else
  result:=exp.ident;
 end
 else
 begin
  result:=target;
  WriteCmd('MOV'+sat(exp),target,exp.ident,'');

 end;
end;

function TPascalShader.WriteExpressionConst(exp:TExpression;const target:string=''):string;
var
 f:single;
begin
 if target='' then
 begin
  if exp.sat then
  begin
   f:=strtofloat(exp.info.tokenstring);
   if f<0 then f:=0;
   if f>1 then f:=1;
   exp.info.tokenstring:=floattostr(f);
  end;
  result:=exp.info.tokenstring;
 end
 else
 begin
  result:=target;
  WriteCmd('MOV'+sat(exp),target,exp.info.tokenstring,'');
 end;
result:='{'+result+','+result+','+result+','+result+'}';
end;


function TPascalShader.WriteExpressionFunc_OpCode1(exp:TExpression;const opcode:string;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat(exp),result,p1,'');
end
else
begin
 WriteCmd(opcode+sat(exp),target,p1,'');
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_OpCode1s(exp:TExpression;const opcode:string;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat(exp),result,p1,'');
end
else
begin
 WriteCmd(opcode+sat(exp),target,p1,'');
 result:=target;
end;
end;


function TPascalShader.WriteExpressionFunc_OpCode2(exp:TExpression;const opcode:string;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat(exp),result,p1,p2);
end
else
begin
 WriteCmd(opcode+sat(exp),target,p1,p2);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_OpCode2s(exp:TExpression;const opcode:string;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
p1:=ChangeScalarSwizzle(p1);
p2:=ChangeScalarSwizzle(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(opcode+sat(exp),result,p1,p2);
end
else
begin
 WriteCmd(opcode+sat(exp),target,p1,p2);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_Normalize(exp:TExpression;const target:string=''):string;
var
 p1,sw:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('MUL'+sat(exp),result+'.xyz',result+'.w',p1);
end
else
begin
 sw:=CombineSwizzle(target,'xyz');
 if sw<>'' then
 begin
  result:=AllocateTempVar;
  WriteCmd('DP3',result+'.w',p1,p1);
  WriteCmd('RSQ',result+'.w',result+'.w','');
  WriteCmd('MUL'+sat(exp),sw,result+'.w',p1);
  FreeTempVar(result);
 end;
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_Dist(exp:TExpression;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ',result+'.w,',result+'.w','');
 WriteCmd('RCP'+sat(exp),result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('RCP'+sat(exp),target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_InvDist(exp:TExpression;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ'+sat(exp),result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3',result+'.w',result,result);
 WriteCmd('RSQ'+sat(exp),target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_SqrDist(exp:TExpression;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3'+sat(exp),result,result,result);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('DP3'+sat(exp),target,result,result);
 FreeTempVar(result);
 result:=target;
end;
end;


function TPascalShader.WriteExpressionFunc_Length(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('RCP'+sat(exp),result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ',result+'.w',result+'.w','');
 WriteCmd('RCP'+sat(exp),target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_SqrLength(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3'+sat(exp),result,p1,p1);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3'+sat(exp),target,p1,p1);
 FreeTempVar(result);
 result:=target;
end;
end;


function TPascalShader.WriteExpressionFunc_InvLength(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ'+sat(exp),result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result+'.w',p1,p1);
 WriteCmd('RSQ'+sat(exp),target,result+'.w','');
 FreeTempVar(result);
 result:=target;
end;
end;


function TPascalShader.WriteExpressionFunc_sqr(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MUL'+sat(exp),result,p1,p1);
end
else
begin
 WriteCmd('MUL'+sat(exp),target,p1,p1);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_sqrt(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('RSQ',result+'.w',p1,'');
 WriteCmd('RCP'+sat(exp),result,result+'.w','');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('RSQ',result+'.w',p1,'');
 WriteCmd('RCP'+sat(exp),target,result+'.w','');

 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_mad(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3:string;
begin
CheckParamCount(exp,3);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
p3:=WriteExpression(exp.paramlist[2]);
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MAD'+sat(exp),result,p1,p2,p3);
end
else
begin
 WriteCmd('MAD'+sat(exp),target,p1,p2,p3);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_MatrixMult3(exp:TExpression;const target:string=''):string;
var
 p1,p2,sw:string;
 row1,row2,row3:string;
begin
CheckParamCount(exp,2);
if exp.paramlist[0].info.token<>tkident then
 raise ESyntaxError.Create(TokenInfoStr(exp.paramlist[0].info)+' Identifier expected ');

p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p2);

if sametext(exp.paramlist[0].ident,'matrix') then
begin
 if length(exp.paramlist[0].paramlist)<>3 then
  raise ESyntaxError.Create(TokenInfoStr(exp.paramlist[0].info)+' Invalid parameter count ');
row1:=WriteExpression(exp.paramlist[0].paramlist[0]);
row2:=WriteExpression(exp.paramlist[0].paramlist[1]);
row3:=WriteExpression(exp.paramlist[0].paramlist[2]);
FreeTempVar(row1);
FreeTempVar(row2);
FreeTempVar(row3);
end
else
begin
p1:=FindValidMatrix(lowercase(exp.paramlist[0].ident));
if p1='' then
 raise ESyntaxError.Create(TokenInfoStr(exp.paramlist[0].info)+' Matrix expected ');
row1:=p1+'.row[0]';
row2:=p1+'.row[1]';
row3:=p1+'.row[2]';
end;

if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3'+sat(exp),result+'.x',row1,p2);
 WriteCmd('DP3'+sat(exp),result+'.y',row2,p2);
 WriteCmd('DP3'+sat(exp),result+'.z',row3,p2);
end
else
begin
 sw:=CombineSwizzle(target,'x');
 if sw<>'' then WriteCmd('DP3'+sat(exp),sw,row1,p2);
  sw:=CombineSwizzle(target,'y');
 if sw<>'' then WriteCmd('DP3'+sat(exp),sw,row2,p2);
  sw:=CombineSwizzle(target,'z');
 if sw<>'' then WriteCmd('DP3'+sat(exp),sw,row3,p2);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_MatrixMult4(exp:TExpression;const target:string=''):string;
var
 p1,p2,sw:string;
 row1,row2,row3,row4:string;
begin
CheckParamCount(exp,2);
if exp.paramlist[0].info.token<>tkident then
 raise ESyntaxError.Create(TokenInfoStr(exp.paramlist[0].info)+' Identifier expected ');

p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p2);

if sametext(exp.paramlist[0].ident,'matrix') then
begin
 if length(exp.paramlist[0].paramlist)<>4 then
  raise ESyntaxError.Create(TokenInfoStr(exp.paramlist[0].info)+' Invalid parameter count ');
row1:=WriteExpression(exp.paramlist[0].paramlist[0]);
row2:=WriteExpression(exp.paramlist[0].paramlist[1]);
row3:=WriteExpression(exp.paramlist[0].paramlist[2]);
row4:=WriteExpression(exp.paramlist[0].paramlist[3]);
FreeTempVar(row1);
FreeTempVar(row2);
FreeTempVar(row3);
FreeTempVar(row4);
end
else
begin
p1:=FindValidMatrix(lowercase(exp.paramlist[0].ident));
if p1='' then
 raise ESyntaxError.Create(TokenInfoStr(exp.paramlist[0].info)+' Matrix expected ');
row1:=p1+'.row[0]';
row2:=p1+'.row[1]';
row3:=p1+'.row[2]';
row4:=p1+'.row[3]';
end;

if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP4'+sat(exp),result+'.x',row1,p2);
 WriteCmd('DP4'+sat(exp),result+'.y',row2,p2);
 WriteCmd('DP4'+sat(exp),result+'.z',row3,p2);
 WriteCmd('DP4'+sat(exp),result+'.w',row4,p2);
end
else
begin
 sw:=CombineSwizzle(target,'x');
 if sw<>'' then WriteCmd('DP4'+sat(exp),sw,row1,p2);
  sw:=CombineSwizzle(target,'y');
 if sw<>'' then WriteCmd('DP4'+sat(exp),sw,row2,p2);
  sw:=CombineSwizzle(target,'z');
 if sw<>'' then WriteCmd('DP4'+sat(exp),sw,row3,p2);
  sw:=CombineSwizzle(target,'w');
 if sw<>'' then WriteCmd('DP4'+sat(exp),sw,row4,p2);

 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_Vector(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3,p4,s,sw:string;
 c,i:integer;
 bl:array[0..3] of boolean;
begin
CheckParamCount(exp,4);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
p3:=WriteExpression(exp.paramlist[2]);
p4:=WriteExpression(exp.paramlist[3]);
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
FreeTempVar(p4);
s:='{';
c:=0;
for i:=0 to 3 do
begin
 if i>0 then s:=s+',';
 if exp.paramlist[i].info.token in [tkinteger,tkfloat] then
 begin
  bl[i]:=false;
  inc(c);
  s:=s+exp.paramlist[i].info.tokenstring;
 end
 else
 begin
  s:=s+'0';
  bl[i]:=true;
 end;
end;
s:=s+'}';
if target='' then
begin
 if c=4 then
 begin
  result:=s;
 end
 else
 begin
  result:=AllocateTempVar;
 if c>0 then
  WriteCmd('MOV'+sat(exp),result,s,'');
 if bl[0] then WriteCmd('MOV'+sat(exp),result+'.x',p1,'');
 if bl[1] then WriteCmd('MOV'+sat(exp),result+'.y',p2,'');
 if bl[2] then WriteCmd('MOV'+sat(exp),result+'.z',p3,'');
 if bl[3] then WriteCmd('MOV'+sat(exp),result+'.w',p4,'');
 end;
end
else
begin
 if c>0 then
  WriteCmd('MOV'+sat(exp),target,s,'');
 sw:=CombineSwizzle(target,'x');
 if (sw<>'') and bl[0] then WriteCmd('MOV'+sat(exp),sw,p1,'');
 sw:=CombineSwizzle(target,'y');
 if (sw<>'') and bl[1] then WriteCmd('MOV'+sat(exp),sw,p2,'');
 sw:=CombineSwizzle(target,'z');
 if (sw<>'') and bl[2] then WriteCmd('MOV'+sat(exp),sw,p3,'');
 sw:=CombineSwizzle(target,'w');
 if (sw<>'') and bl[3] then WriteCmd('MOV'+sat(exp),sw,p4,'');
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_sat(exp:TExpression;const target:string=''):string;
var
 p1:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,'1');
 WriteCmd('MAX',result,result,'0');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,'1');
 WriteCmd('MAX',target,result,'0');
 FreeTempVar(result);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_clamp(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3:string;
begin
CheckParamCount(exp,3);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
p3:=WriteExpression(exp.paramlist[2]);
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,p3);
 WriteCmd('MAX'+sat(exp),result,result,p2);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('MIN',result,p1,p3);
 WriteCmd('MAX'+sat(exp),target,result,p2);
 FreeTempVar(result);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_reflect(exp:TExpression;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result,p1,p2);
 WriteCmd('MUL',result,result,p2);
 WriteCmd('MAD'+sat(exp),result,result,'-2',p1);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('DP3',result,p1,p2);
 WriteCmd('MUL',result,result,p2);
 WriteCmd('MAD'+sat(exp),target,result,'-2',p1);
 FreeTempVar(result);
 result:=target;
end;
end;

function TPascalShader.WriteExpressionFunc_lightatten(exp:TExpression;const target:string=''):string;
var
 p1,p2:string;
begin
CheckParamCount(exp,2);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p1);
FreeTempVar(p2);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('MUL',result,p1,p2);
 WriteCmd('DP3',result,result,result);
 WriteCmd('ADD'+sat(exp),result,'{1,1,1,1}',NegateStr(result));
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('MUL',result,p1,p2);
 WriteCmd('DP3',result,result,result);
 WriteCmd('ADD'+sat(exp),target,'{1,1,1,1}',NegateStr(result));
 FreeTempVar(result);
 result:=target;
end;
end;


function TPascalShader.WriteExpressionFunc(exp:TExpression;const target:string=''):string;
var
 fname:string;
begin
result:='';
fname:=lowercase(exp.ident);
if fname='min' then result:=WriteExpressionFunc_OpCode2(exp,'MIN',target) else
if fname='max' then result:=WriteExpressionFunc_OpCode2(exp,'MAX',target) else
if fname='abs' then result:=WriteExpressionFunc_OpCode1(exp,'ABS',target) else
if fname='cross' then result:=WriteExpressionFunc_OpCode2(exp,'XPD',target) else
if (fname='dot') or (fname='dot3') then result:=WriteExpressionFunc_OpCode2(exp,'DP3',target) else
if fname='dot4' then result:=WriteExpressionFunc_OpCode2(exp,'DP4',target) else
if fname='doth' then result:=WriteExpressionFunc_OpCode2(exp,'DPH',target) else
if fname='floor' then result:=WriteExpressionFunc_OpCode1(exp,'FLR',target) else
if fname='frac' then result:=WriteExpressionFunc_OpCode1(exp,'FRC',target) else
if fname='exp2' then result:=WriteExpressionFunc_OpCode1s(exp,'EX2',target) else
if fname='log2' then result:=WriteExpressionFunc_OpCode1s(exp,'LG2',target) else
if fname='power' then result:=WriteExpressionFunc_OpCode2s(exp,'POW',target) else
if fname='rcp' then result:=WriteExpressionFunc_OpCode1s(exp,'RCP',target) else
if (fname='rsq') or (fname='rsqrt') then result:=WriteExpressionFunc_OpCode1s(exp,'RSQ',target) else
if fname='normalize' then result:=WriteExpressionFunc_Normalize(exp,target) else
if fname='length' then result:=WriteExpressionFunc_Length(exp,target) else
if fname='sqrlength' then result:=WriteExpressionFunc_SqrLength(exp,target) else
if fname='invlength' then result:=WriteExpressionFunc_InvLength(exp,target) else
if fname='dist' then result:=WriteExpressionFunc_Dist(exp,target) else
if fname='sqrdist' then result:=WriteExpressionFunc_SqrDist(exp,target) else
if fname='invdist' then result:=WriteExpressionFunc_InvDist(exp,target) else
if fname='sqr' then result:=WriteExpressionFunc_sqr(exp,target) else
if fname='sqrt' then result:=WriteExpressionFunc_sqrt(exp,target) else
if fname='mad' then result:=WriteExpressionFunc_mad(exp,target) else
if fname='matrixmult3' then result:=WriteExpressionFunc_MatrixMult3(exp,target) else
if fname='matrixmult4' then result:=WriteExpressionFunc_MatrixMult4(exp,target) else
if fname='vector' then result:=WriteExpressionFunc_Vector(exp,target) else
if fname='sat' then result:=WriteExpressionFunc_sat(exp,target) else
if fname='clamp' then result:=WriteExpressionFunc_clamp(exp,target) else
if fname='lightatten' then result:=WriteExpressionFunc_lightatten(exp,target) else
if fname='reflect' then result:=WriteExpressionFunc_reflect(exp,target)
 else result:=WriteExpressionFunction(exp,target);
if result='' then
 raise ESyntaxError.Create(TokenInfoStr(exp.info)+'Undeclared identifier: "'+exp.ident+'"');
end;

function TPascalShader.WriteExpressionFunction(exp:TExpression;const target:string=''):string;
begin
result:='';
end;

function TPascalShader.WriteExpression(exp:TExpression;target:string=''):string;
begin
case exp.info.token of
 tkident:
  begin
   if exp.paramlist=nil then
     result:=WriteExpressionIdent(exp,target)
   else
    result:=WriteExpressionFunc(exp,target);
  end;
 tkinteger,tkfloat:result:=WriteExpressionConst(exp,target);
 tkadd,tksub,tkmult,tkdiv,tkand,
 tkor,tkxor,tkequal,tknotequal,tknot,
 tkgequal,tkgreater,tkless,tklequal:
  begin
   result:=WriteOperator(exp,target);
  end;
end;
end;

procedure TPascalShader.WriteAssignment(st:TStatement);
var
 p1,p2:string;
begin
CheckIdent(st.exp1);
if not (st.exp1.symbolinfo.scope in [scvar,scoutput]) then
 raise ESyntaxError.Create(TokenInfoStr(st.info)+'Left side cannot be assigned to');
st.exp2:=OptimizeExpression(st.exp2);

if currentcondition='' then
begin
 p1:=WriteExpression(st.exp2,WriteExpression(st.exp1));
 FreeTempVar(p1);  // wirklich ntig?
end
else
begin // Ausdruck: x=x+...
 if (st.exp2.info.token=tkadd) and (st.exp2.subnodes[0].info.token=tkident) and sametext(st.exp2.subnodes[0].ident,st.exp1.ident) then
 begin
  p1:=WriteExpression(st.exp1);
  p2:=WriteExpression(st.exp2.subnodes[1]);
  WriteCmd('MAD'+sat(st.exp2),p1,p2,currentcondition,p1);
  end
 else // Ausdruck: x=...-x
 if (st.exp2.info.token=tkadd) and (st.exp2.subnodes[1].info.token=tkident) and sametext(st.exp2.subnodes[1].ident,st.exp1.ident) then
 begin
  p1:=WriteExpression(st.exp1);
  p2:=WriteExpression(st.exp2.subnodes[0]);
  WriteCmd('MAD'+sat(st.exp2),p1,p2,currentcondition,p1);
  end
 else // Ausdruck: x=x-...
 if (st.exp2.info.token=tksub) and (st.exp2.subnodes[0].info.token=tkident) and sametext(st.exp2.subnodes[0].ident,st.exp1.ident) then
 begin
  p1:=WriteExpression(st.exp1);
  p2:=WriteExpression(st.exp2.subnodes[1]);
  WriteCmd('MAD'+sat(st.exp2),p1,p2,NegateStr(currentcondition),p1);
  end
 else  // Ausdruck: x=...-x
 if (st.exp2.info.token=tksub) and (st.exp2.subnodes[1].info.token=tkident) and sametext(st.exp2.subnodes[1].ident,st.exp1.ident) then
 begin
  p1:=WriteExpression(st.exp1);
  p2:=WriteExpression(st.exp2.subnodes[0]);
  WriteCmd('MAD'+sat(st.exp2),p1,p2,NegateStr(currentcondition),p1);
  end
 else
  begin
   p1:=WriteExpression(st.exp1);
   WriteCmd('MUL',p1,p1,currentconditioninv);
   p2:=WriteExpression(st.exp2);
   WriteCmd('MAD',p1,p2,currentcondition,p1);
   FreeTempVar(p2);
  end;
 end;
end;

procedure TPascalShader.WriteCodeBlock(st:TStatement);
var
 i:integer;
begin
for i:=0 to length(st.statements)-1 do
 WriteStatement(st.statements[i]);
end;

procedure TPascalShader.WriteWithStatement(st:TStatement);
var
 s:string;
begin
s:=FindValidWithDecl(st.withdecl);
if s='' then
 raise ESyntaxError.Create(TokenInfoStr(st.info)+' Undeclard Identifier: "'+st.withdecl+'"');
withlist.Add(s);
WriteStatement(st.statements[0]);
withlist.Delete(withlist.count-1);
end;

procedure TPascalShader.WriteIfStatement(st:TStatement);
var
 cond,sgc,sgcinv,s:string;
begin
cond:=WriteExpression(st.exp1);
sgc:=currentcondition;
sgcinv:=currentconditioninv;


currentconditioninv:=AllocateTempVar;
if sgc<>'' then
begin
 currentcondition:=AllocateTempVar;
 WriteCmd('MIN',currentcondition,sgc,cond);
end
else
 currentcondition:=cond;
WriteCmd('SUB',currentconditioninv,'1',currentcondition);

WriteStatement(st.statements[0]);
if length(st.statements)>1 then
begin
s:=currentcondition;
currentconditioninv:=currentcondition;
currentcondition:=s;
WriteStatement(st.statements[1]);
end;

FreeTempVar(currentcondition);
FreeTempVar(currentconditioninv);
currentcondition:=sgc;
currentconditioninv:=sgc;

end;

procedure TPascalShader.WriteDiscard(st:TStatement);
begin
WriteCmd('KIL','',currentcondition,'','');
end;


procedure TPascalShader.WriteStatement(st:TStatement);
begin
case st._type of
 stassignment:WriteAssignment(st);
 stblock:WriteCodeBlock(st);
 stwith:WriteWithStatement(st);
 stif:WriteIfStatement(st);
 stdiscard:WriteDiscard(st);
end;
end;

procedure TPascalShader.WriteVarBlock;
var
 i:integer;
begin
for i:=0 to symbollist.count-1 do
if (TSymbolInfo(symbollist[i]).scope=scvar) and TSymbolInfo(symbollist[i]).used and not TSymbolInfo(symbollist[i])._default then
 WriteLine('TEMP '+TSymbolInfo(symbollist[i]).name+';');
end;

procedure TPascalShader.WriteVaryingBlock;
var
 c,i:integer;
begin
c:=0;
case VaryingType of
 scoutput:
  for i:=0 to symbollist.count-1 do
  if (TSymbolInfo(symbollist[i]).varying) then
  begin
   TSymbolInfo(symbollist[i]).offset:=c;
  if TSymbolInfo(symbollist[i]).used then
   WriteLine('OUTPUT '+TSymbolInfo(symbollist[i]).name+'=result.texcoord['+inttostr(c)+'];');
   inc(c);
  end;
 scattrib:
  for i:=0 to symbollist.count-1 do
  if (TSymbolInfo(symbollist[i]).varying) then
  begin
   TSymbolInfo(symbollist[i]).offset:=c;
   if TSymbolInfo(symbollist[i]).used then
    WriteLine('ATTRIB '+TSymbolInfo(symbollist[i]).name+'=fragment.texcoord['+inttostr(c)+'];');
   inc(c);
  end;
end;
end;


procedure TPascalShader.WriteParamBlock;
var
 i:integer;
begin
for i:=0 to symbollist.count-1 do
 with TSymbolInfo(symbollist[i]) do
 if (scope=scparam) and not _default and used then
 begin
  if arraysize=0 then
   WriteLine('PARAM '+name+'=program.local['+inttostr(offset)+'];')
  else
  if arraysize>0 then
   WriteLine('PARAM '+name+'=program.local['+inttostr(offset)+'..'+inttostr(offset+arraysize-1)+'];')
  end;
end;


procedure TPascalShader.WriteCommandList;
var
 i:integer;
 s:string;
begin
for i:=0 to length(cmdlist)-1 do
if cmdlist[i].cmd<>'' then
begin
 s:=cmdlist[i].cmd;
 if cmdlist[i].p1<>'' then s:=s+' '+cmdlist[i].p1;
 if cmdlist[i].p2<>'' then
 begin
  if cmdlist[i].p1<>'' then s:=s+',' else s:=s+' '; 
  s:=s+cmdlist[i].p2;
 end; 
 if cmdlist[i].p3<>'' then s:=s+','+cmdlist[i].p3;
 if cmdlist[i].p4<>'' then s:=s+','+cmdlist[i].p4;
 WriteLine(s+';');
end;
end;

function TPascalShader.IsVarRead(const cmd:TARBCommand;const v:string):boolean;
begin
result:=Sametext(ExtractVarname(cmd.p2),v) or
        Sametext(ExtractVarname(cmd.p3),v) or Sametext(ExtractVarname(cmd.p4),v);
end;

function TPascalShader.IsVarWritten(const cmd:TARBCommand;const v:string):boolean;
var
 sw:string;
begin
sw:=GetSwizzleMask(cmd.p1);
result:=(v<>'') and Sametext(ExtractVarname(cmd.p1),v) and ((sw='') or Sametext(sw,'xyzw'));
end;

function TPascalShader.CanVarDiscarded(const vn:string;cmdid:integer):boolean;
var
 i:integer;
 si:TSymbolInfo;
begin
si:=FindSymbol(vn);
if assigned(si) then
begin
 result:=si.scope<>scoutput;
 for i:=cmdid to length(cmdlist)-1 do
 if cmdlist[i].cmd<>'' then
 begin
  if IsVarRead(cmdlist[i],vn) then
  begin
   result:=false;
   break;
  end;
  if IsVarWritten(cmdlist[i],vn) then
  begin
   result:=true;
   break;
  end;
 end;
end
else
 result:=false;
end;

function TPascalShader.RemoveDblAssignments:boolean;
var
 i:integer;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd<>'' then
 begin
  if CanVarDiscarded(ExtractVarname(cmdlist[i].p1),i+1) then
  begin
   cmdlist[i].cmd:='';
   result:=true;
  end;
 end;
end;

function TPascalShader.RemoveAssignments:boolean;
var
 i,j:integer;
 v,v2:string;
begin
result:=false;
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd='MOV' then
 begin
  v:=ExtractVarname(cmdlist[i].p1);
  v2:=ExtractVarname(cmdlist[i].p2);
  if FindSymbol(v).scope<>scvar then continue;
  for j:=i+1 to length(cmdlist)-1 do
  begin
   if ExtractVarname(cmdlist[j].p2)=v then
   begin
    cmdlist[j].p2:=ChangeVarname(cmdlist[j].p2,v2);
    result:=true;
    cmdlist[i].cmd:='';
   end;
   if ExtractVarname(cmdlist[j].p3)=v then
   begin
    cmdlist[j].p3:=ChangeVarname(cmdlist[j].p3,v2);
    result:=true;
    cmdlist[i].cmd:='';
   end;
   if ExtractVarname(cmdlist[j].p4)=v then
   begin
    cmdlist[j].p4:=ChangeVarname(cmdlist[j].p4,v2);
    result:=true;
    cmdlist[i].cmd:='';    
   end;
   if IsVarWritten(cmdlist[j],v) or IsVarWritten(cmdlist[j],v2) then break;
  end;
 end;
end;

procedure TPascalShader.FindUsedSymbols;
var
 i:integer;
 si:TSymbolInfo;
begin
for i:=0 to length(cmdlist)-1 do
 if cmdlist[i].cmd<>'' then
 begin
  si:=FindSymbol(ExtractVarname(cmdlist[i].p1));
  if si<>nil then si.used:=true;
  si:=FindSymbol(ExtractVarname(cmdlist[i].p2));
  if si<>nil then si.used:=true;
  si:=FindSymbol(ExtractVarname(cmdlist[i].p3));
  if si<>nil then si.used:=true;
  si:=FindSymbol(ExtractVarname(cmdlist[i].p4));
  if si<>nil then si.used:=true;
 end;
end;


procedure TPascalShader.OptimizeCommandList;
var
 changed:boolean;
begin
repeat
changed:=false;
changed:=changed or RemoveDblAssignments or RemoveAssignments;
until not changed;
FindUsedSymbols;
end;

procedure TPascalShader.WriteProgram;
begin
WriteStatement(codeblock);
OptimizeCommandList;
WriteLine(Header);
WriteVaryingBlock;
WriteParamBlock;
WriteVarBlock;
WriteCommandList;
WriteLine('END');
end;


procedure TPascalShader.InitDefaultVars;
var
 i,j:integer;
begin
InitMatrixList;
AddDefaultSymbol('state.material.ambient',scparam);
AddDefaultSymbol('state.material.diffuse',scparam);
AddDefaultSymbol('state.material.specular',scparam);
AddDefaultSymbol('state.material.emission',scparam);
AddDefaultSymbol('state.material.shininess',scparam);
AddDefaultSymbol('state.material.front.ambient',scparam);
AddDefaultSymbol('state.material.front.diffuse',scparam);
AddDefaultSymbol('state.material.front.specular',scparam);
AddDefaultSymbol('state.material.front.emission',scparam);
AddDefaultSymbol('state.material.front.shininess',scparam);
AddDefaultSymbol('state.material.back.ambient',scparam);
AddDefaultSymbol('state.material.back.diffuse',scparam);
AddDefaultSymbol('state.material.back.specular',scparam);
AddDefaultSymbol('state.material.back.emission',scparam);
AddDefaultSymbol('state.material.back.shininess',scparam);

for i:=0 to 7 do
begin
 AddDefaultSymbol('state.light['+inttostr(i)+'].ambient',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].diffuse',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].specular',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].position',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].attenuation',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].spot.direction',scparam);
 AddDefaultSymbol('state.light['+inttostr(i)+'].half',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].ambient',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].diffuse',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].specular',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].front.ambient',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].front.diffuse',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].front.specular',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].back.ambient',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].back.diffuse',scparam);
 AddDefaultSymbol('state.lightprod['+inttostr(i)+'].back.specular',scparam);
end;
AddDefaultSymbol('state.lightmodel.ambient',scparam);
AddDefaultSymbol('state.lightmodel.scenecolor',scparam);
AddDefaultSymbol('state.lightmodel.front.scenecolor',scparam);
AddDefaultSymbol('state.lightmodel.back.scenecolor',scparam);

for i:=0 to 7 do
begin
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.s',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.t',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.r',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].eye.q',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.s',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.r',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.r',scparam);
 AddDefaultSymbol('state.texgen['+inttostr(i)+'].object.q',scparam);
end;

AddDefaultSymbol('state.fog.color',scparam);
AddDefaultSymbol('state.fog.params',scparam);

for i:=0 to 5 do
 AddDefaultSymbol('state.clip['+inttostr(i)+'].plane',scparam);

AddDefaultSymbol('state.point.size',scparam);
AddDefaultSymbol('state.point.attenuation',scparam);

for i:=0 to matrixlist.Count-1 do
 for j:=0 to 3 do
  AddDefaultSymbol(matrixlist[i]+'.row['+inttostr(j)+']',scparam);

end;

procedure TPascalShader.InitReservedWords;
begin
with reservedwordlist do
begin
 Add('ABS'); Add('ABS_SAT'); Add('ADD'); Add('ADD_SAT'); Add('ALIAS'); Add('ATTRIB'); Add('CMP'); Add('CMP_SAT'); Add('COS');
 Add('COS_SAT'); Add('DP3'); Add('DP3_SAT'); Add('DP4'); Add('DP4_SAT'); Add('DPH'); Add('DPH_SAT'); Add('DST'); Add('DST_SAT');
 Add('END'); Add('EX2'); Add('EX2_SAT'); Add('FLR'); Add('FLR_SAT'); Add('FRC'); Add('FRC_SAT'); Add('KIL'); Add('LG2');
 Add('LG2_SAT'); Add('LIT'); Add('LIT_SAT'); Add('LRP'); Add('LRP_SAT'); Add('MAD'); Add('MAD_SAT'); Add('MAX'); Add('MAX_SAT');
 Add('MIN'); Add('MIN_SAT'); Add('MOV'); Add('MOV_SAT'); Add('MUL'); Add('MUL_SAT'); Add('OPTION'); Add('OUTPUT'); Add('PARAM');
 Add('POW'); Add('POW_SAT'); Add('RCP'); Add('RCP_SAT'); Add('RSQ'); Add('RSQ_SAT'); Add('SIN'); Add('SIN_SAT'); Add('SCS');
 Add('SCS_SAT'); Add('SGE'); Add('SGE_SAT'); Add('SLT'); Add('SLT_SAT'); Add('SUB'); Add('SUB_SAT'); Add('SWZ'); Add('SWZ_SAT');
 Add('TEMP'); Add('TEX'); Add('TEX_SAT'); Add('TXB'); Add('TXB_SAT'); Add('TXP'); Add('TXP_SAT'); Add('XPD'); Add('XPD_SAT');
 Add('fragment'); Add('program'); Add('result'); Add('state'); Add('texture');
 Add('ADDRESS');  Add('ARL'); Add('EXP');Add('LOG'); Add('vertex');
end;
end;

procedure TPascalShader.InitMatrixList;
var
 i:integer;
begin
matrixlist.Add('state.matrix.modelview');
matrixlist.Add('state.matrix.modelview.inverse');
matrixlist.Add('state.matrix.modelview.transpose');
matrixlist.Add('state.matrix.modelview.invtrans');

for i:=0 to 3 do
begin
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+']');
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+'].inverse');
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+'].transpose');
 matrixlist.Add('state.matrix.modelview['+inttostr(i)+'].invtrans');
end;

matrixlist.Add('state.matrix.projection');
matrixlist.Add('state.matrix.projection.inverse');
matrixlist.Add('state.matrix.projection.transpose');
matrixlist.Add('state.matrix.projection.invtrans');


matrixlist.Add('state.matrix.mvp');
matrixlist.Add('state.matrix.mvp.inverse');
matrixlist.Add('state.matrix.mvp.transpose');
matrixlist.Add('state.matrix.mvp.invtrans');

matrixlist.Add('state.matrix.texture');
matrixlist.Add('state.matrix.texture.inverse');
matrixlist.Add('state.matrix.texture.transpose');
matrixlist.Add('state.matrix.texture.invtrans');

for i:=0 to 7 do
begin
 matrixlist.Add('state.matrix.texture['+inttostr(i)+']');
 matrixlist.Add('state.matrix.texture['+inttostr(i)+'].inverse');
 matrixlist.Add('state.matrix.texture['+inttostr(i)+'].transpose');
 matrixlist.Add('state.matrix.texture['+inttostr(i)+'].invtrans');
end;

matrixlist.Add('state.matrix.program');
matrixlist.Add('state.matrix.program.inverse');
matrixlist.Add('state.matrix.program.transpose');
matrixlist.Add('state.matrix.program.invtrans');

for i:=0 to 31 do
begin
 matrixlist.Add('state.matrix.program['+inttostr(i)+']');
 matrixlist.Add('state.matrix.program['+inttostr(i)+'].inverse');
 matrixlist.Add('state.matrix.program['+inttostr(i)+'].transpose');
 matrixlist.Add('state.matrix.program['+inttostr(i)+'].invtrans');
end;



end;

function TPascalShader.GetSymbolOffset(const v:string):integer;
var
 si:TSymbolInfo;
begin
si:=FindSymbol(v);
if si=nil then result:=-1 else result:=si.offset;
end;


constructor TPascalShader.create(st:TStream);
begin
 inherited create;
 symbollist:=TList.create;
 nodelist:=TList.Create;
 expstack:=TStack.Create;
 matrixlist:=TStringList.create;
 withlist:=TStringList.create;
 reservedwordlist:=TStringList.Create;
 reservedwordlist.CaseSensitive:=false;
 InitReservedWords;
 InitDefaultVars;
 SavedDecimalSeparator:=DecimalSeparator;
 DecimalSeparator:='.';
 self.sc:=TScanner.create(st);
 ParseProgram;
 WriteProgram;
 DecimalSeparator:=SavedDecimalSeparator;
end;

destructor TPascalShader.destroy;
var
 i:integer;
begin
DecimalSeparator:=SavedDecimalSeparator;
for i:=0 to symbollist.count-1 do
 TObject(symbollist[i]).free;
for i:=0 to nodelist.count-1 do
 TObject(nodelist[i]).free;
sc.free;
nodelist.free;
symbollist.free;
expstack.Free;
matrixlist.free;
withlist.Free;
reservedwordlist.free;
inherited;
end;

// TVertexProgram


function TVertexProgram.WriteExpressionFunc_Cos(exp:TExpression;const target:string=''):string;
const
 c:array[0..4,0..3] of string=(
       ('0.0','0.5','1.0','0.0'),
       ('0.25','-9.0','0.75','0.1591549430'),
       ('24.9808039603','-24.9808039603','-60.1458091736','60.1458091736'),
       ('85.4537887573','-85.4537887573','-64.9393539429','64.9393539429'),
       ('19.7392082214','-19.7392082214','-1.0','1.0'));
var
 p1,r0,r1,r2:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
r0:=AllocateTempVar;
r1:=AllocateTempVar;
r2:=AllocateTempVar;
WriteCmd('MUL',r1+'.x','{'+c[1,3]+','+c[1,3]+','+c[1,3]+','+c[1,3]+'}',p1);
WriteCmd('FRC',r1+'.y',r1+'.x','');
WriteCmd('SLT',r2+'.x',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('SGE',r2+'.yz',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('DP3',r2+'.y',r2,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');
WriteCmd('ADD',r0+'.xyz','-'+r1+'.y','{'+c[0,0]+','+c[0,1]+','+c[0,2]+','+c[0,3]+'}');
WriteCmd('MUL',r0,r0,r0);
WriteCmd('MAD',r1,'{'+c[2,0]+','+c[2,1]+','+c[2,0]+','+c[2,1]+'}',r0,'{'+c[1,2]+','+c[2,3]+','+c[2,2]+','+c[2,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,0]+','+c[3,1]+','+c[3,0]+','+c[3,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,2]+','+c[3,3]+','+c[3,2]+','+c[3,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,0]+','+c[4,1]+','+c[4,0]+','+c[4,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');

if target='' then
begin
 result:=r0;
 WriteCmd('DP3'+sat(exp),r0,r1,'-'+r2);
 FreeTempVar(r1);
 FreeTempVar(r2);
end
else
begin
 WriteCmd('DP3'+sat(exp),target,r1,'-'+r2);
 FreeTempVar(r0);
 FreeTempVar(r1);
 FreeTempVar(r2);
 result:=target;
end;
end;

function TVertexProgram.WriteExpressionFunc_sin(exp:TExpression;const target:string=''):string;
const
 c:array[0..4,0..3] of string=(
       ('0.0','0.5','1.0','0.0'),
       ('0.25','-9.0','0.75','0.1591549430'),
       ('24.9808039603','-24.9808039603','-60.1458091736','60.1458091736'),
       ('85.4537887573','-85.4537887573','-64.9393539429','64.9393539429'),
       ('19.7392082214','-19.7392082214','-1.0','1.0'));
var
 p1,r0,r1,r2:string;
begin
CheckParamCount(exp,1);
p1:=WriteExpression(exp.paramlist[0]);
FreeTempVar(p1);
p1:=ChangeScalarSwizzle(p1);
r0:=AllocateTempVar;
r1:=AllocateTempVar;
r2:=AllocateTempVar;
WriteCmd('MAD',r1+'.x','{'+c[1,3]+','+c[1,3]+','+c[1,3]+','+c[1,3]+'}',p1,'{-'+c[1,0]+'}');
WriteCmd('FRC',r1+'.y',r1+'.x','');
WriteCmd('SLT',r2+'.x',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('SGE',r2+'.yz',r1+'.y','{'+c[1,0]+','+c[1,1]+','+c[1,2]+','+c[1,3]+'}');
WriteCmd('DP3',r2+'.y',r2,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');
WriteCmd('ADD',r0+'.xyz','-'+r1+'.y','{'+c[0,0]+','+c[0,1]+','+c[0,2]+','+c[0,3]+'}');
WriteCmd('MUL',r0,r0,r0);
WriteCmd('MAD',r1,'{'+c[2,0]+','+c[2,1]+','+c[2,0]+','+c[2,1]+'}',r0,'{'+c[1,2]+','+c[2,3]+','+c[2,2]+','+c[2,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,0]+','+c[3,1]+','+c[3,0]+','+c[3,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[3,2]+','+c[3,3]+','+c[3,2]+','+c[3,3]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,0]+','+c[4,1]+','+c[4,0]+','+c[4,1]+'}');
WriteCmd('MAD',r1,r1,r0,'{'+c[4,2]+','+c[4,3]+','+c[4,2]+','+c[4,3]+'}');

if target='' then
begin
 result:=r0;
 WriteCmd('DP3'+sat(exp),r0,r1,'-'+r2);
 FreeTempVar(r1);
 FreeTempVar(r2);
end
else
begin
 WriteCmd('DP3'+sat(exp),target,r1,'-'+r2);
 FreeTempVar(r0);
 FreeTempVar(r1);
 FreeTempVar(r2);
 result:=target;
end;
end;

function TVertexProgram.WriteExpressionFunc_Lerp(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3:string;
begin
CheckParamCount(exp,3);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
p3:=WriteExpression(exp.paramlist[2]);
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('MAD'+sat(exp),result,p3,result,p1);
end
else
begin
 result:=AllocateTempVar;
 WriteCmd('SUB',result,p1,p2);
 WriteCmd('MAD'+sat(exp),target,p3,result,p1);
 result:=target;
end;
end;


function TVertexProgram.WriteExpressionFunction(exp:TExpression;const target:string=''):string;
var
 fname:string;
begin
fname:=lowercase(exp.ident);
if fname='cos' then result:=WriteExpressionFunc_Cos(exp,target) else
if fname='sin' then result:=WriteExpressionFunc_sin(exp,target) else
if fname='lerp' then result:=WriteExpressionFunc_lerp(exp,target)
else result:='';
end;


procedure TVertexProgram.InitDefaultVars;
var
 i:integer;
begin
inherited;
AddDefaultSymbol('vertex.position',scattrib);
AddDefaultSymbol('vertex.weight',scattrib);
AddDefaultSymbol('vertex.weight[0]',scattrib);
AddDefaultSymbol('vertex.weight[1]',scattrib);
AddDefaultSymbol('vertex.weight[2]',scattrib);
AddDefaultSymbol('vertex.weight[3]',scattrib);
AddDefaultSymbol('vertex.normal',scattrib);
AddDefaultSymbol('vertex.color',scattrib);
AddDefaultSymbol('vertex.color.primary',scattrib);
AddDefaultSymbol('vertex.color.secondary',scattrib);
AddDefaultSymbol('vertex.fogcoord',scattrib);
AddDefaultSymbol('vertex.texcoord',scattrib);
for i:=0 to 7 do
 AddDefaultSymbol('vertex.texcoord['+inttostr(i)+']',scattrib);
AddDefaultSymbol('vertex.matrixindex',scattrib);
AddDefaultSymbol('vertex.matrixindex[0]',scattrib);
AddDefaultSymbol('vertex.matrixindex[4]',scattrib);
AddDefaultSymbol('vertex.matrixindex[8]',scattrib);
AddDefaultSymbol('vertex.matrixindex[12]',scattrib);

for i:=0 to 15 do
 AddDefaultSymbol('vertex.attrib['+inttostr(i)+']',scattrib);

AddDefaultSymbol('result.position',scoutput);
AddDefaultSymbol('result.color',scoutput);
AddDefaultSymbol('result.color.primary',scoutput);
AddDefaultSymbol('result.color.secondary',scoutput);
AddDefaultSymbol('result.color.front',scoutput);
AddDefaultSymbol('result.color.front.primary',scoutput);
AddDefaultSymbol('result.color.front.secondary',scoutput);
AddDefaultSymbol('result.color.back',scoutput);
AddDefaultSymbol('result.color.back.primary',scoutput);
AddDefaultSymbol('result.color.back.secondary',scoutput);
AddDefaultSymbol('result.fogcoord',scoutput);
AddDefaultSymbol('result.texcoord',scoutput);

for i:=0 to 7 do
 AddDefaultSymbol('result.texcoord['+inttostr(i)+']',scoutput);

end;

constructor TVertexProgram.create(st:TStream);
begin
AllowDiscard:=false;
SAT_Cmd:=false;
Header:='!!ARBvp1.0';
VaryingType:=scoutput;
inherited;
end;

// TFragmentProgram

function TFragmentProgram.WriteExpressionFunc_Lerp(exp:TExpression;const target:string=''):string;
var
 p1,p2,p3:string;
begin
CheckParamCount(exp,3);
p1:=WriteExpression(exp.paramlist[0]);
p2:=WriteExpression(exp.paramlist[1]);
p3:=WriteExpression(exp.paramlist[2]);
FreeTempVar(p1);
FreeTempVar(p2);
FreeTempVar(p3);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd('LRP'+sat(exp),result,p1,p2,p3);
end
else
begin
 WriteCmd('LRP'+sat(exp),target,p1,p2,p3);
 result:=target;
end;
end;

function TFragmentProgram.WriteExpressionFunc_texture(exp:TExpression;const texture:string;const target:string=''):string;
var
 p2,p3:string;
 coord,cmd:string;
begin
if not (length(exp.paramlist) in [2,3]) then
 raise ESyntaxError.Create(tokeninfostr(exp.info)+'Invalid parameter count');
if exp.paramlist[0].info.token<>tkinteger then
 raise ESyntaxError.Create(tokeninfostr(exp.paramlist[0].info)+'Integer expected');

p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p2);

if length(exp.paramlist)=3 then
begin
 coord:=AllocateTempVar;
 p3:=WriteExpression(exp.paramlist[2]);
 FreeTempVar(p3);
 if not SameText(coord,p2) then
  WriteCmd('MOV',coord+'.xyz',p2,'');
 WriteCmd('MOV',coord+'.w',ChangeScalarSwizzle(p3),'');
 cmd:='TXB';
end
else
 begin
 coord:=p2;
 cmd:='TEX';
 end;

FreeTempVar(coord);
if target='' then
 result:=AllocateTempVar
else
 result:=target;
WriteCmd(cmd+Sat(exp),result,coord,'texture['+exp.paramlist[0].info.tokenstring+']',texture);

end;

function TFragmentProgram.WriteExpressionFunc_textureproj(exp:TExpression;const texture:string;const target:string=''):string;
var
 p2:string;
begin
CheckParamCount(exp,2);
if exp.paramlist[0].info.token<>tkinteger then
 raise ESyntaxError.Create(tokeninfostr(exp.paramlist[0].info)+'Integer expected');

p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p2);

if target='' then
 result:=AllocateTempVar
else
 result:=target;
WriteCmd('TXP'+Sat(exp),result,p2,'texture['+exp.paramlist[0].info.tokenstring+']',texture);

end;

function TFragmentProgram.WriteExpressionFunc_normalmap(exp:TExpression;const target:string=''):string;
var
 p2,p3:string;
 coord,cmd:string;
begin
if not (length(exp.paramlist) in [2,3]) then
 raise ESyntaxError.Create(tokeninfostr(exp.info)+'Invalid parameter count');
if exp.paramlist[0].info.token<>tkinteger then
 raise ESyntaxError.Create(tokeninfostr(exp.paramlist[0].info)+'Integer expected');

p2:=WriteExpression(exp.paramlist[1]);
FreeTempVar(p2);

if length(exp.paramlist)=3 then
begin
 coord:=AllocateTempVar;
 p3:=WriteExpression(exp.paramlist[2]);
 FreeTempVar(p3);
 if not SameText(coord,p2) then
  WriteCmd('MOV',coord+'.xyz',p2,'');
 WriteCmd('MOV',coord+'.w',ChangeScalarSwizzle(p3),'');
 cmd:='TXB';
end
else
 begin
 coord:=p2;
 cmd:='TEX';
 end;

FreeTempVar(coord);
if target='' then
begin
 result:=AllocateTempVar;
 WriteCmd(cmd,result,coord,'texture['+exp.paramlist[0].info.tokenstring+']','2D');
 WriteCmd('MAD'+sat(exp),result,result,'{2.0,2.0,2.0,1.0}','{-1.0,-1.0,-1.0,0.0}');
end
else
begin
 result:=AllocateTempVar;
 WriteCmd(cmd,result,coord,'texture['+exp.paramlist[0].info.tokenstring+']','2D');
 WriteCmd('MAD'+sat(exp),target,result,'{2.0,2.0,2.0,1.0}','{-1.0,-1.0,-1.0,0.0}');
 FreeTempVar(result);
 result:=target;
end;

end;


function TFragmentProgram.WriteExpressionFunction(exp:TExpression;const target:string=''):string;
var
 fname:string;
begin
fname:=lowercase(exp.ident);
if fname='cos' then result:=WriteExpressionFunc_OpCode1s(exp,'COS',target) else
if fname='sin' then result:=WriteExpressionFunc_OpCode1s(exp,'SIN',target) else
if fname='texture1d' then result:=WriteExpressionFunc_texture(exp,'1D',target) else
if fname='texture2d' then result:=WriteExpressionFunc_texture(exp,'2D',target) else
if fname='texture3d' then result:=WriteExpressionFunc_texture(exp,'3D',target) else
if fname='texturecubemap' then result:=WriteExpressionFunc_texture(exp,'CUBE',target) else
if fname='texturerect' then result:=WriteExpressionFunc_texture(exp,'RECT',target) else
if fname='texture1dproj' then result:=WriteExpressionFunc_textureproj(exp,'1D',target) else
if fname='texture2dproj' then result:=WriteExpressionFunc_textureproj(exp,'2D',target) else
if fname='texture3dproj' then result:=WriteExpressionFunc_textureproj(exp,'3D',target) else
if fname='texturecubemapproj' then result:=WriteExpressionFunc_textureproj(exp,'CUBE',target) else
if fname='texturerectproj' then result:=WriteExpressionFunc_textureproj(exp,'RECT',target) else
if fname='normalmap' then result:=WriteExpressionFunc_normalmap(exp,target) else
if fname='lerp' then result:=WriteExpressionFunc_lerp(exp,target)
else result:='';
end;



procedure TFragmentProgram.InitDefaultVars;
var
 i:integer;
begin
inherited;
AddDefaultSymbol('fragment.color',scattrib);
AddDefaultSymbol('fragment.color.primary',scattrib);
AddDefaultSymbol('fragment.color.secondary',scattrib);
AddDefaultSymbol('fragment.texcoord',scattrib);
for i:=0 to 7 do
 AddDefaultSymbol('fragment.texcoord['+inttostr(i)+']',scattrib);
AddDefaultSymbol('fragment.fogcoord',scattrib);
AddDefaultSymbol('fragment.position',scattrib);

AddDefaultSymbol('result.color',scoutput);
AddDefaultSymbol('result.depth',scoutput);
end;

constructor TFragmentProgram.create(st:TStream);
begin
AllowDiscard:=true;
SAT_Cmd:=true;
Header:='!!ARBfp1.0';
VaryingType:=scattrib;
inherited;
end;

begin
end.

